import Foundation

protocol CardStackAnimatable {
  func animateReset(_ cardStack: SwipeCardStack, topCard: SwipeCard)
  func animateShift(_ cardStack: SwipeCardStack, withDistance distance: Int)
  func animateSwipe(_ cardStack: SwipeCardStack,
                    topCard: SwipeCard,
                    direction: SwipeDirection,
                    forced: Bool)
  func animateUndo(_ cardStack: SwipeCardStack, topCard: SwipeCard)
  func removeAllCardAnimations(_ cardStack: SwipeCardStack)
  func removeBackgroundCardAnimations(_ cardStack: SwipeCardStack)
}


/// The background card animator for the card stack.
///
/// All methods should be called only after the `CardStackManager` has been updated.
class CardStackAnimator: CardStackAnimatable {

  static let shared = CardStackAnimator()

  // MARK: - Main Methods

  func animateReset(_ cardStack: SwipeCardStack, topCard: SwipeCard) {
    removeBackgroundCardAnimations(cardStack)

    Animator.animateKeyFrames(withDuration: resetDuration(cardStack, topCard: topCard),
                              options: .allowUserInteraction,
                              animations: { [weak self] in
                                self?.addCancelSwipeAnimationKeyFrames(cardStack)
      }, completion: nil)
  }

  func animateSwipe(_ cardStack: SwipeCardStack,
                    topCard: SwipeCard,
                    direction: SwipeDirection,
                    forced: Bool) {
    removeBackgroundCardAnimations(cardStack)

    let delay = swipeDelay(for: topCard, forced: forced)
    let duration = swipeDuration(cardStack,
                                 topCard: topCard,
                                 direction: direction,
                                 forced: forced)

    // no background cards left to animate, so we instead just delay calling the completion block
    if cardStack.visibleCards.count == 0 {
      DispatchQueue.main.asyncAfter(deadline: .now() + delay + duration) {
        cardStack.swipeCompletionBlock()
      }
      return
    }

    Animator.animateKeyFrames(withDuration: duration,
                              delay: delay,
                              animations: { [weak self] in
                                self?.addSwipeAnimationKeyFrames(cardStack)
    }) { finished in
      if finished {
        cardStack.swipeCompletionBlock()
      }
    }
  }

  func animateShift(_ cardStack: SwipeCardStack, withDistance distance: Int) {
    removeAllCardAnimations(cardStack)

    // place background cards in old positions
    for (index, card) in cardStack.visibleCards.enumerated() {
      card.transform = cardStack.transform(forCardAtIndex: index + distance)
    }

    // animate background cards to new positions
    Animator.animateKeyFrames(withDuration: shiftDuration(cardStack),
                              animations: { [weak self] in
                                self?.addShiftAnimationKeyFrames(cardStack)
    }) { finshed in
      if finshed {
        cardStack.shiftCompletionBlock()
      }
    }
  }

  func animateUndo(_ cardStack: SwipeCardStack, topCard: SwipeCard) {
    removeBackgroundCardAnimations(cardStack)

    // place background cards in old positions
    for (index, card) in cardStack.backgroundCards.enumerated() {
      card.transform = cardStack.transform(forCardAtIndex: index)
    }

    // animate background cards to new positions
    Animator.animateKeyFrames(withDuration: undoDuration(cardStack, topCard: topCard),
                              animations: { [weak self] in
                                self?.addUndoAnimationKeyFrames(cardStack)
    }) { finished in
      if finished {
        cardStack.undoCompletionBlock()
      }
    }
  }

  func removeBackgroundCardAnimations(_ cardStack: SwipeCardStack) {
    for card in cardStack.backgroundCards {
      card.removeAllAnimations()
    }
  }

  func removeAllCardAnimations(_ cardStack: SwipeCardStack) {
    for card in cardStack.visibleCards {
      card.removeAllAnimations()
    }
  }

  // MARK: - Animation Keyframes

  func addCancelSwipeAnimationKeyFrames(_ cardStack: SwipeCardStack) -> Void {
    for (index, card) in cardStack.backgroundCards.enumerated() {
      let transform = cardStack.transform(forCardAtIndex: index + 1)
      Animator.addTransformKeyFrame(to: card, transform: transform)
    }
  }

  func addShiftAnimationKeyFrames(_ cardStack: SwipeCardStack) -> Void {
    for (index, card) in cardStack.visibleCards.enumerated() {
      let transform = cardStack.transform(forCardAtIndex: index)
      Animator.addTransformKeyFrame(to: card, transform: transform)
    }
  }

  func addSwipeAnimationKeyFrames(_ cardStack: SwipeCardStack) -> Void {
    for (index, card) in cardStack.visibleCards.enumerated() {
      Animator.addKeyFrame {
        cardStack.layoutCard(card, at: index)
      }
    }
  }

  func addUndoAnimationKeyFrames(_ cardStack: SwipeCardStack) -> Void {
    for (index, card) in cardStack.backgroundCards.enumerated() {
      Animator.addKeyFrame {
        cardStack.layoutCard(card, at: index + 1)
      }
    }
  }

  // MARK: - Animation Calculations

  func resetDuration(_ cardStack: SwipeCardStack, topCard: SwipeCard) -> TimeInterval {
    return cardStack.animationOptions.resetDuration
      ??  topCard.animationOptions.totalResetDuration / 2
  }

  func shiftDuration(_ cardStack: SwipeCardStack) -> TimeInterval {
    return cardStack.animationOptions.shiftDuration
  }

  func swipeDelay(for topCard: SwipeCard, forced: Bool) -> TimeInterval {
    let duration = topCard.animationOptions.totalSwipeDuration
    let relativeOverlayDuration = topCard.animationOptions.relativeSwipeOverlayFadeDuration
    let delay = duration * TimeInterval(relativeOverlayDuration)
    return forced ? delay : 0
  }

  func swipeDuration(_ cardStack: SwipeCardStack,
                     topCard: SwipeCard,
                     direction: SwipeDirection,
                     forced: Bool) -> TimeInterval {
    if let swipeDuration = cardStack.animationOptions.swipeDuration {
      return swipeDuration
    }

    if forced {
      return topCard.animationOptions.totalSwipeDuration / 2
    }

    let velocityFactor = topCard.dragSpeed(on: direction) / topCard.minimumSwipeSpeed(on: direction)

    // card swiped below the minimum swipe speed
    if velocityFactor < 1.0 {
      return topCard.animationOptions.totalSwipeDuration / 2
    }

    // card swiped at least the minimum swipe speed -> return relative duration
    return 1.0 / (2.0 * TimeInterval(velocityFactor))
  }

  func undoDuration(_ cardStack: SwipeCardStack, topCard: SwipeCard) -> TimeInterval {
    return cardStack.animationOptions.undoDuration
      ?? topCard.animationOptions.totalReverseSwipeDuration / 2
  }
}
